@@ -26,18 +26,18 @@
 #include <sys/stat.h>
 #include <fcntl.h>
 #include <unistd.h>
+#include <wacom-properties.h>
 
 /*****************************************************************************
  * wcmAllocate --
  * Allocate the generic bits needed by any wacom device, regardless of type.
  ****************************************************************************/
 
-static int wcmAllocate(LocalDevicePtr local)
+static int wcmAllocate(InputInfoPtr pInfo)
 {
 	WacomDevicePtr   priv   = NULL;
 	WacomCommonPtr   common = NULL;
 	WacomToolPtr     tool   = NULL;
-	WacomToolAreaPtr area   = NULL;
 	int i;
 
 	priv = calloc(1, sizeof(WacomDeviceRec));
@@ -52,28 +52,17 @@ static int wcmAllocate(LocalDevicePtr local)
 	if(!tool)
 		goto error;
 
-	area = calloc(1, sizeof(WacomToolArea));
-	if (!area)
-		goto error;
-
-	local->flags = 0;
-	local->device_control = gWacomModule.DevProc;
-	local->read_input = gWacomModule.DevReadInput;
-	local->control_proc = gWacomModule.DevChangeControl;
-	local->close_proc = gWacomModule.DevClose;
-	local->switch_mode = gWacomModule.DevSwitchMode;
-	local->atom = 0;
-	local->dev = NULL;
-	local->private = priv;
-	local->private_flags = 0;
-	local->old_x = -1;
-	local->old_y = -1;
+	pInfo->device_control = gWacomModule.DevProc;
+	pInfo->read_input = gWacomModule.DevReadInput;
+	pInfo->control_proc = gWacomModule.DevChangeControl;
+	pInfo->switch_mode = gWacomModule.DevSwitchMode;
+	pInfo->dev = NULL;
+	pInfo->private = priv;
 
 	priv->next = NULL;
-	priv->local = local;
+	priv->pInfo = pInfo;
 	priv->common = common;       /* common info pointer */
-	priv->hardProx = 1;	     /* previous hardware proximity */
-	priv->screen_no = -1;        /* associated screen */
+	priv->oldCursorHwProx = 0;   /* previous cursor hardware proximity */
 	priv->nPressCtrl [0] = 0;    /* pressure curve x0 */
 	priv->nPressCtrl [1] = 0;    /* pressure curve y0 */
 	priv->nPressCtrl [2] = 100;  /* pressure curve x1 */
@@ -96,21 +85,14 @@ static int wcmAllocate(LocalDevicePtr local)
 	priv->striprup = 4;			/* Default right strip up event */
 	priv->striprdn = 5;			/* Default right strip down event */
 	priv->naxes = 6;			/* Default number of axes */
-	priv->numScreen = screenInfo.numScreens; /* configured screens count */
-	priv->currentScreen = -1;                /* current screen in display */
-	priv->twinview = TV_NONE;		/* not using twinview gfx */
-	priv->wcmMMonitor = 1;			/* enabled (=1) to support multi-monitor desktop. */
-						/* disabled (=0) when user doesn't want to move the */
-						/* cursor from one screen to another screen */
 
 	/* JEJ - throttle sampling code */
 	priv->throttleLimit = -1;
 
-	common->wcmFlags = RAW_FILTERING_FLAG;   /* various flags */
+	common->wcmFlags = 0;               /* various flags */
 	common->wcmDevices = priv;
-	common->wcmProtocolLevel = 4;      /* protocol level */
-	common->wcmTPCButton = 
-		common->wcmTPCButtonDefault; /* set Tablet PC button on/off */
+	common->wcmProtocolLevel = WCM_PROTOCOL_4; /* protocol level */
+	common->wcmTPCButton = 0;          /* set Tablet PC button on/off */
 	common->wcmCapacity = -1;          /* Capacity is disabled */
 	common->wcmCapacityDefault = -1;    /* default to -1 when capacity isn't supported */
 					   /* 3 when capacity is supported */
@@ -141,18 +123,12 @@ static int wcmAllocate(LocalDevicePtr local)
 	priv->tool = tool;
 	common->wcmTool = tool;
 	tool->next = NULL;          /* next tool in list */
-	tool->arealist = area;      /* list of defined areas */
+	tool->device = pInfo;
 	/* tool->typeid is set once we know the type - see wcmSetType */
 
-	/* tool area */
-	priv->toolarea = area;
-	area->next = NULL;    /* next area in list */
-	area->device = local; /* associated WacomDevice */
-
 	return 1;
 
 error:
-	free(area);
 	free(tool);
 	wcmFreeCommon(&common);
 	free(priv);
@@ -164,37 +140,31 @@ error:
  * Free the memory allocated by wcmAllocate
  ****************************************************************************/
 
-static void wcmFree(LocalDevicePtr local)
+static void wcmFree(InputInfoPtr pInfo)
 {
-	WacomDevicePtr priv = local->private;
+	WacomDevicePtr priv = pInfo->private;
 
 	if (!priv)
 		return;
 
-	free(priv->toolarea);
 	free(priv->tool);
 	wcmFreeCommon(&priv->common);
 	free(priv);
 
-	local->private = NULL;
+	pInfo->private = NULL;
 }
 
-static int wcmSetType(LocalDevicePtr local, const char *type)
+static int wcmSetType(InputInfoPtr pInfo, const char *type)
 {
-	WacomDevicePtr priv = local->private;
+	WacomDevicePtr priv = pInfo->private;
 
 	if (!type)
-	{
-		xf86Msg(X_ERROR, "%s: No type or invalid type specified.\n"
-				"Must be one of stylus, touch, cursor, eraser, or pad\n",
-				local->name);
-		return 0;
-	}
+		goto invalid;
 
 	if (xf86NameCmp(type, "stylus") == 0)
 	{
 		priv->flags = ABSOLUTE_FLAG|STYLUS_ID;
-		local->type_name = XI_STYLUS;
+		pInfo->type_name = WACOM_PROP_XI_TYPE_STYLUS;
 	} else if (xf86NameCmp(type, "touch") == 0)
 	{
 		int flags = TOUCH_ID;
@@ -203,20 +173,21 @@ static int wcmSetType(LocalDevicePtr local, const char *type)
 			flags |= ABSOLUTE_FLAG;
 
 		priv->flags = flags;
-		local->type_name = XI_TOUCH;
+		pInfo->type_name = WACOM_PROP_XI_TYPE_TOUCH;
 	} else if (xf86NameCmp(type, "cursor") == 0)
 	{
 		priv->flags = CURSOR_ID;
-		local->type_name = XI_CURSOR;
+		pInfo->type_name = WACOM_PROP_XI_TYPE_CURSOR;
 	} else if (xf86NameCmp(type, "eraser") == 0)
 	{
 		priv->flags = ABSOLUTE_FLAG|ERASER_ID;
-		local->type_name = XI_ERASER;
+		pInfo->type_name = WACOM_PROP_XI_TYPE_ERASER;
 	} else if (xf86NameCmp(type, "pad") == 0)
 	{
-		priv->flags = PAD_ID;
-		local->type_name = XI_PAD;
-	}
+		priv->flags = ABSOLUTE_FLAG|PAD_ID;
+		pInfo->type_name = WACOM_PROP_XI_TYPE_PAD;
+	} else
+		goto invalid;
 
 	/* Set the device id of the "last seen" device on this tool */
 	priv->old_device_id = wcmGetPhyDeviceID(priv);
@@ -227,6 +198,12 @@ static int wcmSetType(LocalDevicePtr local, const char *type)
 	priv->tool->typeid = DEVICE_ID(priv->flags); /* tool type (stylus/touch/eraser/cursor/pad) */
 
 	return 1;
+
+invalid:
+	xf86Msg(X_ERROR, "%s: No type or invalid type specified.\n"
+			 "Must be one of stylus, touch, cursor, eraser, or pad\n",
+			 pInfo->name);
+	return 0;
 }
 
 int wcmGetPhyDeviceID(WacomDevicePtr priv)
@@ -252,7 +229,7 @@ int wcmGetPhyDeviceID(WacomDevicePtr priv)
  * starts making sense again.
  */
 
-static const char *default_options[] =
+static char *default_options[] =
 {
 	"StopBits",    "1",
 	"DataBits",    "8",
@@ -265,15 +242,17 @@ static const char *default_options[] =
 
 /* wcmUninit - called when the device is no longer needed. */
 
-static void wcmUninit(InputDriverPtr drv, LocalDevicePtr local, int flags)
+static void wcmUninit(InputDriverPtr drv, InputInfoPtr pInfo, int flags)
 {
-	WacomDevicePtr priv = (WacomDevicePtr) local->private;
+	WacomDevicePtr priv = (WacomDevicePtr) pInfo->private;
 	WacomDevicePtr dev;
 	WacomDevicePtr *prev;
 	WacomCommonPtr common = priv->common;
 
 	DBG(1, priv, "\n");
 
+	/* Server 1.10 will UnInit all devices for us */
+#if GET_ABI_MAJOR(ABI_XINPUT_VERSION) < 12
 	if (priv->isParent)
 	{
 		/* HAL removal sees the parent device removed first. */
@@ -281,7 +260,7 @@ static void wcmUninit(InputDriverPtr drv, LocalDevicePtr local, int flags)
 		dev = priv->common->wcmDevices;
 
 		xf86Msg(X_INFO, "%s: removing automatically added devices.\n",
-			local->name);
+			pInfo->name);
 
 		while(dev)
 		{
@@ -289,31 +268,16 @@ static void wcmUninit(InputDriverPtr drv, LocalDevicePtr local, int flags)
 			if (!dev->isParent)
 			{
 				xf86Msg(X_INFO, "%s: removing dependent device '%s'\n",
-					local->name, dev->local->name);
-				DeleteInputDeviceRequest(dev->local->dev);
+					pInfo->name, dev->pInfo->name);
+				DeleteInputDeviceRequest(dev->pInfo->dev);
 			}
 			dev = next;
 		}
 
-		free(local->name);
-		local->name = NULL;
-	}
-
-	if (priv->toolarea)
-	{
-		WacomToolAreaPtr *prev_area = &priv->tool->arealist;
-		WacomToolAreaPtr area = *prev_area;
-		while (area)
-		{
-			if (area == priv->toolarea)
-			{
-				*prev_area = area->next;
-				break;
-			}
-			prev_area = &area->next;
-			area = area->next;
-		}
+		free(pInfo->name);
+		pInfo->name = NULL;
 	}
+#endif
 
 	if (priv->tool)
 	{
@@ -344,8 +308,8 @@ static void wcmUninit(InputDriverPtr drv, LocalDevicePtr local, int flags)
 		dev = dev->next;
 	}
 
-	wcmFree(local);
-	xf86DeleteInput(local, 0);
+	wcmFree(pInfo);
+	xf86DeleteInput(pInfo, 0);
 }
 
 /* wcmMatchDevice - locate matching device and merge common structure. If an
@@ -357,11 +321,11 @@ static void wcmUninit(InputDriverPtr drv, LocalDevicePtr local, int flags)
  * Returns 1 on a found match or 0 otherwise.
  * Common_return is set to the common struct in use by this device.
  */
-static Bool wcmMatchDevice(LocalDevicePtr pLocal, WacomCommonPtr *common_return)
+static Bool wcmMatchDevice(InputInfoPtr pLocal, WacomCommonPtr *common_return)
 {
 	WacomDevicePtr priv = (WacomDevicePtr)pLocal->private;
 	WacomCommonPtr common = priv->common;
-	LocalDevicePtr pMatch = xf86FirstLocalDevice();
+	InputInfoPtr pMatch = xf86FirstLocalDevice();
 
 	*common_return = common;
 
@@ -378,6 +342,11 @@ static Bool wcmMatchDevice(LocalDevicePtr pLocal, WacomCommonPtr *common_return)
 		{
 			DBG(2, priv, "port share between %s and %s\n",
 					pLocal->name, pMatch->name);
+			/* FIXME: we loose the common->wcmTool here but it
+			 * gets re-added during wcmParseOptions. This is
+			 * currently required by the code, adding the tool
+			 * again here means we trigger the duplicate tool
+			 * detection */
 			wcmFreeCommon(&priv->common);
 			priv->common = wcmRefCommon(privMatch->common);
 			priv->next = priv->common->wcmDevices;
@@ -395,67 +364,146 @@ static Bool wcmMatchDevice(LocalDevicePtr pLocal, WacomCommonPtr *common_return)
  * behind the fd.
  */
 static Bool
-wcmDetectDeviceClass(const LocalDevicePtr local)
+wcmDetectDeviceClass(const InputInfoPtr pInfo)
 {
-	WacomDevicePtr priv = (WacomDevicePtr)local->private;
+	WacomDevicePtr priv = (WacomDevicePtr)pInfo->private;
 	WacomCommonPtr common = priv->common;
 
 	if (common->wcmDevCls)
 		return TRUE;
 
 	/* Bluetooth is also considered as USB */
-	if (gWacomISDV4Device.Detect(local))
+	if (gWacomISDV4Device.Detect(pInfo))
 		common->wcmDevCls = &gWacomISDV4Device;
-	else if (gWacomUSBDevice.Detect(local))
+	else if (gWacomUSBDevice.Detect(pInfo))
 		common->wcmDevCls = &gWacomUSBDevice;
 	else
-		xf86Msg(X_ERROR, "%s: cannot identify device class.\n", local->name);
+		xf86Msg(X_ERROR, "%s: cannot identify device class.\n", pInfo->name);
 
 	return (common->wcmDevCls != NULL);
 }
 
 static Bool
-wcmInitModel(LocalDevicePtr local)
+wcmInitModel(InputInfoPtr pInfo)
 {
-	WacomDevicePtr priv = (WacomDevicePtr)local->private;
+	WacomDevicePtr priv = (WacomDevicePtr)pInfo->private;
 	WacomCommonPtr common = priv->common;
 	char id[BUFFER_SIZE];
 	float version;
 
 	/* Initialize the tablet */
-	if(common->wcmDevCls->Init(local, id, &version) != Success ||
-		wcmInitTablet(local, id, version) != Success)
+	if(common->wcmDevCls->Init(pInfo, id, &version) != Success ||
+		wcmInitTablet(pInfo, id, version) != Success)
 		return FALSE;
 
 	return TRUE;
 }
 
+/**
+ * Link the touch tool to the pen of the same device
+ * so we can arbitrate the events when posting them.
+ */
+static void wcmLinkTouchAndPen(InputInfoPtr pInfo)
+{
+	WacomDevicePtr priv = pInfo->private;
+	WacomCommonPtr common = priv->common;
+	InputInfoPtr device = xf86FirstLocalDevice();
+	WacomCommonPtr tmpcommon = NULL;
+	WacomDevicePtr tmppriv = NULL;
+	Bool touch_device_assigned = FALSE;
+
+	/* Lookup to find the associated pen and touch */
+	for (; device != NULL; device = device->next)
+	{
+		if (!strcmp(device->drv->driverName, "wacom"))
+		{
+			tmppriv = (WacomDevicePtr) device->private;
+			tmpcommon = tmppriv->common;
+			touch_device_assigned = (common->wcmTouchDevice ||
+						tmpcommon->wcmTouchDevice);
+
+			/* skip the same tool or already linked devices */
+			if ((tmppriv == priv) || touch_device_assigned)
+				continue;
+
+			if (tmpcommon->tablet_id == common->tablet_id)
+			{
+				if (IsTouch(tmppriv) && IsPen(priv))
+					common->wcmTouchDevice = tmppriv;
+				else if (IsTouch(priv) && IsPen(tmppriv))
+					tmpcommon->wcmTouchDevice = priv;
+
+				if (common->wcmTouchDevice ||
+						tmpcommon->wcmTouchDevice)
+				{
+					common->tablet_type |= WCM_PENTOUCH;
+					tmpcommon->tablet_type |= WCM_PENTOUCH;
+				}
+			}
+		}
+	}
+}
+
+/**
+ * Check if this device was hotplugged by the driver by checking the _source
+ * option.
+ *
+ * Must be called before wcmNeedAutoHotplug()
+ *
+ * @return True if the source for this device is the wacom driver itself or
+ * false otherwise.
+ */
+static int wcmIsHotpluggedDevice(InputInfoPtr pInfo)
+{
+	char *source = xf86CheckStrOption(pInfo->options, "_source", "");
+	return !strcmp(source, "_driver/wacom");
+}
+
 /* wcmPreInit - called for each input devices with the driver set to
  * "wacom" */
-static LocalDevicePtr wcmPreInit(InputDriverPtr drv, IDevPtr dev, int flags)
+#if GET_ABI_MAJOR(ABI_XINPUT_VERSION) < 12
+static int NewWcmPreInit(InputDriverPtr drv, InputInfoPtr pInfo, int flags);
+
+static InputInfoPtr wcmPreInit(InputDriverPtr drv, IDevPtr dev, int flags)
 {
-	LocalDevicePtr local = NULL;
-	WacomDevicePtr priv = NULL;
-	WacomCommonPtr common = NULL;
-	const char*	type;
-	char*		device, *oldname;
-	int		need_hotplug = 0;
+	InputInfoPtr pInfo = NULL;
 
-	gWacomModule.wcmDrv = drv;
+	if (!(pInfo = xf86AllocateInput(drv, 0)))
+		return NULL;
 
-	if (!(local = xf86AllocateInput(drv, 0)))
-		goto SetupProc_fail;
-
-	local->conf_idev = dev;
-	local->name = dev->identifier;
+	pInfo->conf_idev = dev;
+	pInfo->name = dev->identifier;
 
 	/* Force default port options to exist because the init
 	 * phase is based on those values.
 	 */
-	xf86CollectInputOptions(local, default_options, NULL);
+	xf86CollectInputOptions(pInfo, (const char**)default_options, NULL);
+	xf86ProcessCommonOptions(pInfo, pInfo->options);
 
-	device = xf86SetStrOption(local->options, "Device", NULL);
-	type = xf86FindOptionValue(local->options, "Type");
+	if (NewWcmPreInit(drv, pInfo, flags) == Success) {
+		pInfo->flags |= XI86_CONFIGURED;
+		return pInfo;
+	} else {
+		xf86DeleteInput(pInfo, 0);
+		return NULL;
+	}
+}
+
+static int NewWcmPreInit(InputDriverPtr drv, InputInfoPtr pInfo, int flags)
+#else
+static int wcmPreInit(InputDriverPtr drv, InputInfoPtr pInfo, int flags)
+#endif
+{
+	WacomDevicePtr priv = NULL;
+	WacomCommonPtr common = NULL;
+	const char*	type;
+	char*		device, *oldname;
+	int		need_hotplug = 0, is_dependent = 0;
+
+	gWacomModule.wcmDrv = drv;
+
+	device = xf86SetStrOption(pInfo->options, "Device", NULL);
+	type = xf86FindOptionValue(pInfo->options, "Type");
 
 	/*
 	   Init process:
@@ -469,92 +517,91 @@ static LocalDevicePtr wcmPreInit(InputDriverPtr drv, IDevPtr dev, int flags)
 	   - hotplug dependent devices if needed
 	 */
 
-	if (!wcmAllocate(local))
+	if (!wcmAllocate(pInfo))
 		goto SetupProc_fail;
 
-	if (!device && !(device = wcmEventAutoDevProbe(local)))
+	if (!device && !(device = wcmEventAutoDevProbe(pInfo)))
 		goto SetupProc_fail;
 
-	priv = (WacomDevicePtr) local->private;
+	priv = (WacomDevicePtr) pInfo->private;
 	priv->common->device_path = device;
-	priv->name = local->name;
+	priv->name = pInfo->name;
 
 	/* check if the same device file has been added already */
-	if (wcmIsDuplicate(device, local))
+	if (wcmIsDuplicate(device, pInfo))
 		goto SetupProc_fail;
 
-	if (wcmOpen(local) != Success)
+	if (wcmOpen(pInfo) != Success)
 		goto SetupProc_fail;
 
 	/* Try to guess whether it's USB or ISDV4 */
-	if (!wcmDetectDeviceClass(local))
+	if (!wcmDetectDeviceClass(pInfo))
 		goto SetupProc_fail;
 
 	/* check if this is the first tool on the port */
-	if (!wcmMatchDevice(local, &common))
+	if (!wcmMatchDevice(pInfo, &common))
 		/* initialize supported keys with the first tool on the port */
-		wcmDeviceTypeKeys(local);
+		wcmDeviceTypeKeys(pInfo);
 
-	oldname = local->name;
+	oldname = pInfo->name;
 
-	if ((need_hotplug = wcmNeedAutoHotplug(local, &type)))
+	if (wcmIsHotpluggedDevice(pInfo))
+		is_dependent = 1;
+	else if ((need_hotplug = wcmNeedAutoHotplug(pInfo, &type)))
 	{
 		/* we need subdevices, change the name so all of them have a
 		   type. */
-		char *new_name = Xprintf("%s %s", local->name, type);
-		local->name = priv->name = new_name;
+		char *new_name;
+		if (asprintf(&new_name, "%s %s", pInfo->name, type) == -1)
+			new_name = strdup(pInfo->name);
+		pInfo->name = priv->name = new_name;
 	}
 
 	/* check if the type is valid for those don't need hotplug */
-	if(!need_hotplug && !wcmIsAValidType(local, type))
+	if(!need_hotplug && !wcmIsAValidType(pInfo, type))
 		goto SetupProc_fail;
 
-	if (!wcmSetType(local, type))
+	if (!wcmSetType(pInfo, type))
 		goto SetupProc_fail;
 
-	/* Process the common options. */
-	xf86ProcessCommonOptions(local, local->options);
-	if (!wcmParseOptions(local, need_hotplug))
+	if (!wcmParseOptions(pInfo, need_hotplug, is_dependent))
 		goto SetupProc_fail;
 
-	if (!wcmInitModel(local))
+	if (!wcmInitModel(pInfo))
 		goto SetupProc_fail;
 
-	/* mark the device configured */
-	local->flags |= XI86_POINTER_CAPABLE | XI86_CONFIGURED;
-
 	if (need_hotplug)
 	{
 		priv->isParent = 1;
-		wcmHotplugOthers(local, oldname);
+		wcmHotplugOthers(pInfo, oldname);
 	}
 
-	if (local->fd != -1)
+	if (pInfo->fd != -1)
 	{
-		close(local->fd);
-		local->fd = -1;
+		close(pInfo->fd);
+		pInfo->fd = -1;
 	}
 
-	return (local);
+	/* only link them once per port. We need to try for both pen and touch
+	 * since we do not know which tool (touch or pen) will be added first.
+	 */
+	if (IsTouch(priv) || (IsPen(priv) && !common->wcmTouchDevice))
+		wcmLinkTouchAndPen(pInfo);
+
+	return Success;
 
 SetupProc_fail:
 	/* restart the device list from the next one */
 	if (common && priv)
 		common->wcmDevices = priv->next;
 
-	if (local)
+	if (pInfo && pInfo->fd != -1)
 	{
-		if (local->fd != -1)
-		{
-			close(local->fd);
-			local->fd = -1;
-		}
-
-		wcmFree(local);
-		xf86DeleteInput(local, 0);
+		close(pInfo->fd);
+		pInfo->fd = -1;
 	}
 
-	return NULL;
+	return BadMatch;
 }
 
 InputDriverRec WACOM =
@@ -565,7 +612,9 @@ InputDriverRec WACOM =
 	wcmPreInit,    /* pre-init */
 	wcmUninit, /* un-init */
 	NULL,          /* module */
-	0              /* ref count */
+#if GET_ABI_MAJOR(ABI_XINPUT_VERSION) >= 12
+	default_options
+#endif
 };
 
 
@@ -605,4 +654,4 @@ _X_EXPORT XF86ModuleData wacomModuleData =
 	wcmUnplug
 };
 
-/* vim: set noexpandtab shiftwidth=8: */
+/* vim: set noexpandtab tabstop=8 shiftwidth=8: */
