@@ -1,6 +1,6 @@
 /*
  * Copyright 1995-2002 by Frederic Lepied, France. <Lepied@XFree86.org>
- * Copyright 2002-2010 by Ping Cheng, Wacom. <pingc@wacom.com>
+ * Copyright 2002-2011 by Ping Cheng, Wacom. <pingc@wacom.com>
  * 
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License
@@ -28,9 +28,14 @@
 #include "isdv4.h"
 #include <unistd.h>
 #include <fcntl.h>
+#include <libudev.h>
 
 #define RESET_RELATIVE(ds) do { (ds).relwheel = 0; } while (0)
 
+/* resolution in points/m */
+#define ISDV4_PEN_RESOLUTION    100000
+#define ISDV4_TOUCH_RESOLUTION  10000
+
 /* ISDV4 init process
    This process is the same for other backends (i.e. USB).
 
@@ -53,7 +58,6 @@
 
  */
 
-
 typedef struct {
 	/* Counter for dependent devices. We can only send one QUERY command to
 	   the tablet and we must not send the SAMPLING command until the last
@@ -64,18 +68,18 @@ typedef struct {
 	int baudrate;
 } wcmISDV4Data;
 
-static Bool isdv4Detect(LocalDevicePtr);
-static Bool isdv4ParseOptions(LocalDevicePtr local);
-static Bool isdv4Init(LocalDevicePtr, char* id, float *version);
-static int isdv4ProbeKeys(LocalDevicePtr local);
+static Bool isdv4Detect(InputInfoPtr);
+static Bool isdv4ParseOptions(InputInfoPtr pInfo);
+static Bool isdv4Init(InputInfoPtr, char* id, float *version);
+static int isdv4ProbeKeys(InputInfoPtr pInfo);
 static void isdv4InitISDV4(WacomCommonPtr, const char* id, float version);
-static int isdv4GetRanges(LocalDevicePtr);
-static int isdv4StartTablet(LocalDevicePtr);
-static int isdv4StopTablet(LocalDevicePtr);
-static int isdv4Parse(LocalDevicePtr, const unsigned char* data, int len);
-static int wcmSerialValidate(LocalDevicePtr local, const unsigned char* data);
-static int wcmWaitForTablet(LocalDevicePtr local, char * data, int size);
-static int wcmWriteWait(LocalDevicePtr local, const char* request);
+static int isdv4GetRanges(InputInfoPtr);
+static int isdv4StartTablet(InputInfoPtr);
+static int isdv4StopTablet(InputInfoPtr);
+static int isdv4Parse(InputInfoPtr, const unsigned char* data, int len);
+static int wcmSerialValidate(InputInfoPtr pInfo, const unsigned char* data);
+static int wcmWaitForTablet(InputInfoPtr pInfo, char * data, int size);
+static int wcmWriteWait(InputInfoPtr pInfo, const char* request);
 
 	WacomDeviceClass gWacomISDV4Device =
 	{
@@ -95,10 +99,10 @@ static int wcmWriteWait(LocalDevicePtr local, const char* request);
 		isdv4Parse,
 	};
 
-static void memdump(LocalDevicePtr local, char *buffer, unsigned int len)
+static void memdump(InputInfoPtr pInfo, char *buffer, unsigned int len)
 {
 #ifdef DEBUG
-	WacomDevicePtr priv = (WacomDevicePtr)local->private;
+	WacomDevicePtr priv = (WacomDevicePtr)pInfo->private;
 	WacomCommonPtr common = priv->common;
 	int i;
 
@@ -149,9 +153,9 @@ static int wcmSkipInvalidBytes(const unsigned char* data, int len)
  *   positive number of bytes to skip on error.
  ****************************************************************************/
 
-static int wcmSerialValidate(LocalDevicePtr local, const unsigned char* data)
+static int wcmSerialValidate(InputInfoPtr pInfo, const unsigned char* data)
 {
-	WacomDevicePtr priv = (WacomDevicePtr)local->private;
+	WacomDevicePtr priv = (WacomDevicePtr)pInfo->private;
 	WacomCommonPtr common = priv->common;
 
 	int n;
@@ -163,7 +167,7 @@ static int wcmSerialValidate(LocalDevicePtr local, const unsigned char* data)
 		int n = wcmSkipInvalidBytes(data, common->wcmPktLength);
 		xf86Msg(X_WARNING,
 			"%s: missing header bit. skipping %d bytes.\n",
-			local->name, n);
+			pInfo->name, n);
 		return n;
 	}
 
@@ -174,7 +178,7 @@ static int wcmSerialValidate(LocalDevicePtr local, const unsigned char* data)
 	n = wcmSkipInvalidBytes(&data[1], common->wcmPktLength - 1);
 	n += 1; /* the header byte we already checked */
 	if (n != common->wcmPktLength) {
-		xf86Msg(X_WARNING, "%s: bad data at %d v=%x l=%d\n", local->name,
+		xf86Msg(X_WARNING, "%s: bad data at %d v=%x l=%d\n", pInfo->name,
 			n, data[n], common->wcmPktLength);
 		return n;
 	}
@@ -186,12 +190,12 @@ static int wcmSerialValidate(LocalDevicePtr local, const unsigned char* data)
  * isdv4Detect -- Test if the attached device is ISDV4.
  ****************************************************************************/
 
-static Bool isdv4Detect(LocalDevicePtr local)
+static Bool isdv4Detect(InputInfoPtr pInfo)
 {
 	struct serial_struct ser;
 	int rc;
 
-	rc = ioctl(local->fd, TIOCGSERIAL, &ser);
+	rc = ioctl(pInfo->fd, TIOCGSERIAL, &ser);
 	if (rc == -1)
 		return FALSE;
 
@@ -201,9 +205,9 @@ static Bool isdv4Detect(LocalDevicePtr local)
 /*****************************************************************************
  * isdv4ParseOptions -- parse ISDV4-specific options
  ****************************************************************************/
-static Bool isdv4ParseOptions(LocalDevicePtr local)
+static Bool isdv4ParseOptions(InputInfoPtr pInfo)
 {
-	WacomDevicePtr priv = (WacomDevicePtr)local->private;
+	WacomDevicePtr priv = (WacomDevicePtr)pInfo->private;
 	WacomCommonPtr common = priv->common;
 	wcmISDV4Data *isdv4data;
 	int baud;
@@ -211,19 +215,19 @@ static Bool isdv4ParseOptions(LocalDevicePtr local)
 	/* Determine default baud rate */
 	baud = (common->tablet_id == 0x90)? 19200 : 38400;
 
-	baud = xf86SetIntOption(local->options, "BaudRate", baud);
+	baud = xf86SetIntOption(pInfo->options, "BaudRate", baud);
 
 	switch (baud)
 	{
 		case 38400:
 		case 19200:
 			/* xf86OpenSerial() takes the baud rate from the options */
-			xf86ReplaceIntOption(local->options, "BaudRate", baud);
+			xf86ReplaceIntOption(pInfo->options, "BaudRate", baud);
 			break;
 		default:
 			xf86Msg(X_ERROR, "%s: Illegal speed value "
 					"(must be 19200 or 38400).",
-					local->name);
+					pInfo->name);
 			return FALSE;
 	}
 
@@ -232,7 +236,7 @@ static Bool isdv4ParseOptions(LocalDevicePtr local)
 		if (!(common->private = calloc(1, sizeof(wcmISDV4Data))))
 		{
 			xf86Msg(X_ERROR, "%s: failed to alloc backend-specific data.\n",
-				local->name);
+				pInfo->name);
 			return FALSE;
 		}
 		isdv4data = common->private;
@@ -248,16 +252,16 @@ static Bool isdv4ParseOptions(LocalDevicePtr local)
  * isdv4Init --
  ****************************************************************************/
 
-static Bool isdv4Init(LocalDevicePtr local, char* id, float *version)
+static Bool isdv4Init(InputInfoPtr pInfo, char* id, float *version)
 {
-	WacomDevicePtr priv = (WacomDevicePtr)local->private;
+	WacomDevicePtr priv = (WacomDevicePtr)pInfo->private;
 	WacomCommonPtr common = priv->common;
 	wcmISDV4Data *isdv4data = common->private;
 
 	DBG(1, priv, "initializing ISDV4 tablet\n");
 
 	/* Set baudrate */
-	if (xf86SetSerialSpeed(local->fd, isdv4data->baudrate) < 0)
+	if (xf86SetSerialSpeed(pInfo->fd, isdv4data->baudrate) < 0)
 		return !Success;
 
 	if(id)
@@ -275,30 +279,30 @@ static Bool isdv4Init(LocalDevicePtr local, char* id, float *version)
  * isdv4Query -- Query the device
  ****************************************************************************/
 
-static int isdv4Query(LocalDevicePtr local, const char* query, char* data)
+static int isdv4Query(InputInfoPtr pInfo, const char* query, char* data)
 {
 #ifdef DEBUG
-	WacomDevicePtr priv = (WacomDevicePtr)local->private;
+	WacomDevicePtr priv = (WacomDevicePtr)pInfo->private;
 #endif
 
 	DBG(1, priv, "Querying ISDV4 tablet\n");
 
-	if (isdv4StopTablet(local) != Success)
+	if (isdv4StopTablet(pInfo) != Success)
 		return !Success;
 
 	/* Send query command to the tablet */
-	if (!wcmWriteWait(local, query))
+	if (!wcmWriteWait(pInfo, query))
 		return !Success;
 
 	/* Read the control data */
-	if (!wcmWaitForTablet(local, data, ISDV4_PKGLEN_TPCCTL))
+	if (!wcmWaitForTablet(pInfo, data, ISDV4_PKGLEN_TPCCTL))
 		return !Success;
 
 	/* Control data bit check */
 	if ( !(data[0] & 0x40) )
 	{
 		/* Reread the control data since it may fail the first time */
-		wcmWaitForTablet(local, data, ISDV4_PKGLEN_TPCCTL);
+		wcmWaitForTablet(pInfo, data, ISDV4_PKGLEN_TPCCTL);
 		if ( !(data[0] & 0x40) )
 			return !Success;
 	}
@@ -312,15 +316,13 @@ static int isdv4Query(LocalDevicePtr local, const char* query, char* data)
 
 static void isdv4InitISDV4(WacomCommonPtr common, const char* id, float version)
 {
-	/* set parameters */
-	common->wcmProtocolLevel = 4;
 	/* length of a packet */
 	common->wcmPktLength = ISDV4_PKGLEN_TPCPEN;
 
-	/* digitizer X resolution in points/inch */
-	common->wcmResolX = 2540; 	
-	/* digitizer Y resolution in points/inch */
-	common->wcmResolY = 2540; 	
+	/* digitizer X resolution in points/m */
+	common->wcmResolX = ISDV4_PEN_RESOLUTION;
+	/* digitizer Y resolution in points/m */
+	common->wcmResolY = ISDV4_PEN_RESOLUTION;
 
 	/* tilt disabled */
 	common->wcmFlags &= ~TILT_ENABLED_FLAG;
@@ -330,10 +332,10 @@ static void isdv4InitISDV4(WacomCommonPtr common, const char* id, float version)
  * isdv4GetRanges -- get ranges of the device
  ****************************************************************************/
 
-static int isdv4GetRanges(LocalDevicePtr local)
+static int isdv4GetRanges(InputInfoPtr pInfo)
 {
 	char data[BUFFER_SIZE];
-	WacomDevicePtr priv = (WacomDevicePtr)local->private;
+	WacomDevicePtr priv = (WacomDevicePtr)pInfo->private;
 	WacomCommonPtr common =	priv->common;
 	wcmISDV4Data *isdv4data = common->private;
 	int ret = Success;
@@ -344,14 +346,14 @@ static int isdv4GetRanges(LocalDevicePtr local)
 		goto out;
 
 	/* Set baudrate to configured value */
-	if (xf86SetSerialSpeed(local->fd, isdv4data->baudrate) < 0)
+	if (xf86SetSerialSpeed(pInfo->fd, isdv4data->baudrate) < 0)
 	{
 		ret = !Success;
 		goto out;
 	}
 
 	/* Send query command to the tablet */
-	ret = isdv4Query(local, ISDV4_QUERY, data);
+	ret = isdv4Query(pInfo, ISDV4_QUERY, data);
 	if (ret != Success)
 	{
 		int baud;
@@ -360,20 +362,20 @@ static int isdv4GetRanges(LocalDevicePtr local)
 		baud = (isdv4data->baudrate == 38400)? 19200 : 38400;
 
 		xf86Msg(X_WARNING, "%s: Query failed with %d baud. Trying %d.\n",
-				   local->name, isdv4data->baudrate, baud);
+				   pInfo->name, isdv4data->baudrate, baud);
 
-		if (xf86SetSerialSpeed(local->fd, baud) < 0)
+		if (xf86SetSerialSpeed(pInfo->fd, baud) < 0)
 		{
 			ret = !Success;
 			goto out;
 		}
 
-		ret = isdv4Query(local, ISDV4_QUERY, data);
+		ret = isdv4Query(pInfo, ISDV4_QUERY, data);
 
 		if (ret == Success) {
 			isdv4data->baudrate = baud;
 			/* xf86OpenSerial() takes the baud rate from the options */
-			xf86ReplaceIntOption(local->options, "BaudRate", baud);
+			xf86ReplaceIntOption(pInfo->options, "BaudRate", baud);
 		}
 
 	}
@@ -387,12 +389,12 @@ static int isdv4GetRanges(LocalDevicePtr local)
 		if (rc <= 0)
 		{
 			xf86Msg(X_ERROR, "%s: Error while parsing ISDV4 query.\n",
-					local->name);
+					pInfo->name);
 			if (rc == 0)
 				DBG(2, common, "reply or len invalid.\n");
 			else
 				DBG(2, common, "header data corrupt.\n");
-			memdump(local, data, sizeof(reply));
+			memdump(pInfo, data, sizeof(reply));
 			ret = BadAlloc;
 			goto out;
 		}
@@ -422,7 +424,7 @@ static int isdv4GetRanges(LocalDevicePtr local)
 
 	/* Touch might be supported. Send a touch query command */
 	if (isdv4data->baudrate == 38400 &&
-	    isdv4Query(local, ISDV4_TOUCH_QUERY, data) == Success)
+	    isdv4Query(pInfo, ISDV4_TOUCH_QUERY, data) == Success)
 	{
 		ISDV4TouchQueryReply reply;
 		int rc;
@@ -431,12 +433,12 @@ static int isdv4GetRanges(LocalDevicePtr local)
 		if (rc <= 0)
 		{
 			xf86Msg(X_ERROR, "%s: Error while parsing ISDV4 touch query.\n",
-					local->name);
+					pInfo->name);
 			if (rc == 0)
 				DBG(2, common, "reply or len invalid.\n");
 			else
 				DBG(2, common, "header data corrupt.\n");
-			memdump(local, data, sizeof(reply));
+			memdump(pInfo, data, sizeof(reply));
 			ret = BadAlloc;
 			goto out;
 		}
@@ -482,7 +484,7 @@ static int isdv4GetRanges(LocalDevicePtr local)
 				{
 				    xf86Msg(X_WARNING, "%s: tablet id(%x)"
 					    " mismatch with data id (0x01) \n",
-					    local->name, common->tablet_id);
+					    pInfo->name, common->tablet_id);
 				    goto out;
 				}
 				break;
@@ -493,7 +495,7 @@ static int isdv4GetRanges(LocalDevicePtr local)
 				{
 				    xf86Msg(X_WARNING, "%s: tablet id(%x)"
 					    " mismatch with data id (0x03) \n",
-					    local->name, common->tablet_id);
+					    pInfo->name, common->tablet_id);
 				    goto out;
 				}
 				break;
@@ -510,7 +512,7 @@ static int isdv4GetRanges(LocalDevicePtr local)
 				(1 << reply.panel_resolution);
 
 		if (reply.panel_resolution)
-			common->wcmTouchResolX = common->wcmTouchResolY = 10;
+			common->wcmTouchResolX = common->wcmTouchResolY = ISDV4_TOUCH_RESOLUTION;
 
 		common->wcmVersion = reply.version;
 		ret = Success;
@@ -522,7 +524,7 @@ static int isdv4GetRanges(LocalDevicePtr local)
 			common->wcmTouchResolY);
 	}
 
-	xf86Msg(X_INFO, "%s: serial tablet id 0x%X.\n", local->name, common->tablet_id);
+	xf86Msg(X_INFO, "%s: serial tablet id 0x%X.\n", pInfo->name, common->tablet_id);
 
 out:
 	if (ret == Success)
@@ -534,9 +536,9 @@ out:
 	return ret;
 }
 
-static int isdv4StartTablet(LocalDevicePtr local)
+static int isdv4StartTablet(InputInfoPtr pInfo)
 {
-	WacomDevicePtr priv = (WacomDevicePtr)local->private;
+	WacomDevicePtr priv = (WacomDevicePtr)pInfo->private;
 	WacomCommonPtr common =	priv->common;
 	wcmISDV4Data *isdv4data = common->private;
 
@@ -544,22 +546,22 @@ static int isdv4StartTablet(LocalDevicePtr local)
 		return Success;
 
 	/* Tell the tablet to start sending coordinates */
-	if (!wcmWriteWait(local, ISDV4_SAMPLING))
+	if (!wcmWriteWait(pInfo, ISDV4_SAMPLING))
 		return !Success;
 
 	return Success;
 }
 
-static int isdv4StopTablet(LocalDevicePtr local)
+static int isdv4StopTablet(InputInfoPtr pInfo)
 {
 #if DEBUG
-	WacomDevicePtr priv = (WacomDevicePtr)local->private;
+	WacomDevicePtr priv = (WacomDevicePtr)pInfo->private;
 	WacomCommonPtr common = priv->common;
 #endif
 	int fd_flags;
 
 	/* Send stop command to the tablet */
-	if (!wcmWriteWait(local, ISDV4_STOP))
+	if (!wcmWriteWait(pInfo, ISDV4_STOP))
 		return !Success;
 
 	/* Wait 250 mSecs */
@@ -567,21 +569,21 @@ static int isdv4StopTablet(LocalDevicePtr local)
 		return !Success;
 
 	/* discard potential data on the line */
-	fd_flags = fcntl(local->fd, F_GETFL);
-	if (fcntl(local->fd, F_SETFL, fd_flags | O_NONBLOCK) == 0)
+	fd_flags = fcntl(pInfo->fd, F_GETFL);
+	if (fcntl(pInfo->fd, F_SETFL, fd_flags | O_NONBLOCK) == 0)
 	{
 		char buffer[10];
-		while (read(local->fd, buffer, sizeof(buffer)) > 0)
+		while (read(pInfo->fd, buffer, sizeof(buffer)) > 0)
 			DBG(10, common, "discarding garbage data.\n");
-		fcntl(local->fd, F_SETFL, fd_flags);
+		fcntl(pInfo->fd, F_SETFL, fd_flags);
 	}
 
 	return Success;
 }
 
-static int isdv4Parse(LocalDevicePtr local, const unsigned char* data, int len)
+static int isdv4Parse(InputInfoPtr pInfo, const unsigned char* data, int len)
 {
-	WacomDevicePtr priv = (WacomDevicePtr)local->private;
+	WacomDevicePtr priv = (WacomDevicePtr)pInfo->private;
 	WacomCommonPtr common = priv->common;
 	WacomDeviceState* last = &common->wcmChannel[0].valid.state;
 	WacomDeviceState* lastTemp = &common->wcmChannel[1].valid.state;
@@ -635,7 +637,7 @@ static int isdv4Parse(LocalDevicePtr local, const unsigned char* data, int len)
 	/* Coordinate data bit check */
 	if (data[0] & CONTROL_BIT) /* control data */
 		return common->wcmPktLength;
-	else if ((n = wcmSerialValidate(local,data)) > 0)
+	else if ((n = wcmSerialValidate(pInfo,data)) > 0)
 		return n;
 
 	/* pick up where we left off, minus relative values */
@@ -651,7 +653,7 @@ static int isdv4Parse(LocalDevicePtr local, const unsigned char* data, int len)
 		if (rc == -1)
 		{
 			xf86Msg(X_ERROR, "%s: failed to parse touch data.\n",
-				local->name);
+				pInfo->name);
 			return 0;
 		}
 
@@ -706,7 +708,7 @@ static int isdv4Parse(LocalDevicePtr local, const unsigned char* data, int len)
 
 		if (rc == -1)
 		{
-			xf86Msg(X_ERROR, "%s: failed to parse coordinate data.\n", local->name);
+			xf86Msg(X_ERROR, "%s: failed to parse coordinate data.\n", pInfo->name);
 			return 0;
 		}
 
@@ -768,18 +770,18 @@ static int isdv4Parse(LocalDevicePtr local, const unsigned char* data, int len)
  *   send a request
  ****************************************************************************/
 
-static int wcmWriteWait(LocalDevicePtr local, const char* request)
+static int wcmWriteWait(InputInfoPtr pInfo, const char* request)
 {
 	int len, maxtry = MAXTRY;
 
 	/* send request string */
 	do
 	{
-		len = xf86WriteSerial(local->fd, request, strlen(request));
+		len = xf86WriteSerial(pInfo->fd, request, strlen(request));
 		if ((len == -1) && (errno != EAGAIN))
 		{
 			xf86Msg(X_ERROR, "%s: wcmWriteWait error : %s\n",
-					local->name, strerror(errno));
+					pInfo->name, strerror(errno));
 			return 0;
 		}
 
@@ -789,7 +791,7 @@ static int wcmWriteWait(LocalDevicePtr local, const char* request)
 
 	if (!maxtry)
 		xf86Msg(X_WARNING, "%s: Failed to issue command '%s' "
-				   "after %d tries.\n", local->name, request, MAXTRY);
+				   "after %d tries.\n", pInfo->name, request, MAXTRY);
 
 	return maxtry;
 }
@@ -799,20 +801,20 @@ static int wcmWriteWait(LocalDevicePtr local, const char* request)
  *   wait for tablet data
  ****************************************************************************/
 
-static int wcmWaitForTablet(LocalDevicePtr local, char* answer, int size)
+static int wcmWaitForTablet(InputInfoPtr pInfo, char* answer, int size)
 {
 	int len, maxtry = MAXTRY;
 
 	/* Read size bytes of the answer */
 	do
 	{
-		if ((len = xf86WaitForInput(local->fd, 1000000)) > 0)
+		if ((len = xf86WaitForInput(pInfo->fd, 1000000)) > 0)
 		{
-			len = xf86ReadSerial(local->fd, answer, size);
+			len = xf86ReadSerial(pInfo->fd, answer, size);
 			if ((len == -1) && (errno != EAGAIN))
 			{
 				xf86Msg(X_ERROR, "%s: xf86ReadSerial error : %s\n",
-						local->name, strerror(errno));
+						pInfo->name, strerror(errno));
 				return 0;
 			}
 		}
@@ -822,69 +824,26 @@ static int wcmWaitForTablet(LocalDevicePtr local, char* answer, int size)
 	if (!maxtry)
 		xf86Msg(X_WARNING, "%s: Waited too long for answer "
 				   "(failed after %d tries).\n",
-				   local->name, MAXTRY);
+				   pInfo->name, MAXTRY);
 
 	return maxtry;
 }
 
-/**
- * Query the device's fd for the key bits and the tablet ID. Returns the ID
- * on success or 0 on failure.
- * For serial devices, we set the BTN_TOOL_DOUBLETAP etc. bits based on the
- * device ID. This matching only works for wacom devices (serial ID of
- * WACf), all others are simply assumed to be pen + erasor.
- */
-static int isdv4ProbeKeys(LocalDevicePtr local)
+static int set_keybits_wacom(int id, unsigned long *keys)
 {
-	int id;
 	int tablet_id = 0;
-	struct serial_struct tmp;
-	const char *device = xf86SetStrOption(local->options, "Device", NULL);
-	WacomDevicePtr  priv = (WacomDevicePtr)local->private;
-	WacomCommonPtr  common = priv->common;
-
-	if (ioctl(local->fd, TIOCGSERIAL, &tmp) < 0)
-		return 0;
-
-	/* check device name for ID first */
-	if (sscanf(local->name, "WACf%x", &id) <= 1)
-	{
-		/* id in file sys/class/tty/%str/device/id */
-		FILE *file;
-		char sysfs_id[256];
-		char *str = strstr(device, "ttyS");
-		snprintf(sysfs_id, sizeof(sysfs_id),
-				"/sys/class/tty/%s/device/id", str);
-		file = fopen(sysfs_id, "r");
-
-		/* return true since it falls to default */
-		if (file)
-		{
-			/* make sure we fall to default */
-			if (fscanf(file, "WACf%x\n", &id) <= 0)
-				id = 0;
-
-			fclose(file);
-		}
-	}
-
-	memset(common->wcmKeys, 0, sizeof(common->wcmKeys));
-
-	/* default to penabled */
-	SETBIT(common->wcmKeys, BTN_TOOL_PEN);
-	SETBIT(common->wcmKeys, BTN_TOOL_RUBBER);
 
 	/* id < 0x008 are only penabled */
 	if (id > 0x007)
-		SETBIT(common->wcmKeys, BTN_TOOL_DOUBLETAP);
+		SETBIT(keys, BTN_TOOL_FINGER);
 	if (id > 0x0a)
-		SETBIT(common->wcmKeys, BTN_TOOL_TRIPLETAP);
+		SETBIT(keys, BTN_TOOL_DOUBLETAP);
 
 	/* no pen 2FGT */
 	if (id == 0x010)
 	{
-		CLEARBIT(common->wcmKeys, BTN_TOOL_PEN);
-		CLEARBIT(common->wcmKeys, BTN_TOOL_RUBBER);
+		CLEARBIT(keys, BTN_TOOL_PEN);
+		CLEARBIT(keys, BTN_TOOL_RUBBER);
 	}
 
 	/* 0x9a and 0x9f are only detected by communicating
@@ -903,4 +862,140 @@ static int isdv4ProbeKeys(LocalDevicePtr local)
 	return tablet_id;
 }
 
-/* vim: set noexpandtab shiftwidth=8: */
+static int set_keybits_fujitsu(int id, unsigned long *keys)
+{
+	int tablet_id = 0x90; /* default to penabled */
+
+	if (id == 0x2e7) {
+		SETBIT(keys, BTN_TOOL_DOUBLETAP);
+		tablet_id = 0xe3;
+	}
+
+	if (id == 0x2e9) {
+		SETBIT(keys, BTN_TOOL_FINGER);
+		tablet_id = 0x93;
+	}
+
+	return tablet_id;
+}
+
+typedef struct {
+	const char *pattern; /* sscanf matching pattern to extract ID */
+	/* set the bits in the given keys array based on the id. return the
+	 * tablet_id or the closest guess anyway */
+	int (*set_bits)(int id, unsigned long* keys);
+} ISDV4ModelDesc;
+
+static ISDV4ModelDesc isdv4_models[] = {
+	{ "WACf%x", set_keybits_wacom },
+	{ "FUJ%x", set_keybits_fujitsu },
+	{ NULL, 0 }
+};
+
+/**
+ * Query the model number from the sysfs /sys/.../device/id file and return
+ * the matching model and the ID for the model we found.
+ *
+ * @param pInfo Used for debugging purposes only.
+ * @param fd File descriptor to the serial device.
+ * @param[out] id On success, returns the numeric ID for this device
+ * according to the model-specific matching pattern.
+ *
+ * @return The model description for the matching device or NULL if no
+ * matching one could be found.
+ */
+static ISDV4ModelDesc*
+model_from_sysfs(const InputInfoPtr pInfo, int fd, int *id)
+{
+	WacomDevicePtr priv = pInfo->private;
+	ISDV4ModelDesc* model = NULL;
+	struct udev *udev = NULL;
+	struct udev_device *device = NULL;
+	struct stat st;
+	char *sysfs_path = NULL;
+	FILE *file = NULL;
+
+	fstat(fd, &st);
+
+	udev = udev_new();
+	device = udev_device_new_from_devnum(udev, 'c', st.st_rdev);
+
+	if (!device)
+		goto out;
+
+	if (asprintf(&sysfs_path, "%s/device/id",
+		     udev_device_get_syspath(device)) == -1)
+		goto out;
+
+	DBG(8, priv, "sysfs path: %s\n", sysfs_path);
+
+	file = fopen(sysfs_path, "r");
+	if (!file)
+		goto out;
+
+	model = isdv4_models;
+
+	while(model->pattern && fscanf(file, model->pattern, id) <= 0)
+		model++;
+
+	if (!model->pattern)
+		model = NULL;
+
+	DBG(8, priv, "sysfs check found %s:%d\n",
+	    (model) ? model->pattern : "<unknown>", *id);
+
+out:
+	udev_device_unref(device);
+	udev_unref(udev);
+	fclose(file);
+	free(sysfs_path);
+	return model;
+}
+
+/**
+ * Query the device's fd for the key bits and the tablet ID. Returns the ID
+ * on success. If the model vendor is unknown, we assume a penabled device
+ * (0x90). If the model vendor is known but the model itself is unknown, the
+ * return value depends on the model-specific matching code (0 for Wacom,
+ * 0x90 for Fujitsu).
+ *
+ * For serial devices, we set the BTN_TOOL_DOUBLETAP etc. bits based on the
+ * device ID. This matching only works for known devices (see the
+ * isdv4_model list), all others are simply assumed to be pen + erasor.
+ */
+static int isdv4ProbeKeys(InputInfoPtr pInfo)
+{
+	int id = 0;
+	int tablet_id = 0x90;
+	struct serial_struct tmp;
+	WacomDevicePtr  priv = (WacomDevicePtr)pInfo->private;
+	WacomCommonPtr  common = priv->common;
+	ISDV4ModelDesc *model = isdv4_models;
+
+	if (ioctl(pInfo->fd, TIOCGSERIAL, &tmp) < 0)
+		return 0;
+
+	/* check device name for ID first */
+	while (model->pattern && sscanf(pInfo->name, model->pattern, &id) < 1)
+		model++;
+
+	/* grab id from sysfs/.../device/id */
+	if (!model->pattern)
+		model = model_from_sysfs(pInfo, pInfo->fd, &id);
+
+	memset(common->wcmKeys, 0, sizeof(common->wcmKeys));
+
+	/* default to penabled */
+	SETBIT(common->wcmKeys, BTN_TOOL_PEN);
+	SETBIT(common->wcmKeys, BTN_TOOL_RUBBER);
+
+	if (model && model->set_bits)
+		tablet_id = model->set_bits(id, common->wcmKeys);
+
+	/* Change to generic protocol to match USB MT format */
+	common->wcmProtocolLevel = WCM_PROTOCOL_GENERIC;
+
+	return tablet_id;
+}
+
+/* vim: set noexpandtab tabstop=8 shiftwidth=8: */
